/*!
 * © 2021 David Vines
 *
 * Creative Commons License
 *
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 */

/* globals ROT, Result */
/* exported OpeningMove, Contination, Bout */
class OpeningMove {
	constructor(name) {
		this._name = name;
	}
	toString() {
		return this._name;
	}
	getOdds(theirMove,us,them) {
		throw new Error("getOdds not provided for "+this.toString()+".getOdds("+theirMove.toString()+","+us.toString()+","+them.toString()+")");
	}
}
OpeningMove.Charge = new OpeningMove("Charge");
OpeningMove.Charge.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === OpeningMove.Charge:
			odds = 61; break;
		case theirMove === OpeningMove.Attack:
			odds = 35; break;
		case theirMove === OpeningMove.StepAside:
			odds = 80; break;
		case theirMove === OpeningMove.Slap:
			odds = 46; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for OpeningMove.Charge");
	}
	odds += (us.getSpeed() - them.getWeight())/2.5;
	return odds;
};
OpeningMove.Charge.bothSameOpeningMoveNeedingRestart = function(result,us,them) {
	result.addText("Both "+us.toString()+" and "+them.toString()+" charge.",false);
	return false;
};
OpeningMove.Charge.addSuccessfulDescription = function(result,us,them) {
	result.addText(us.toString()+" charges "+them.toString()+".");
};
OpeningMove.Charge.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" charges "+them.toString()+" pushing "+them.toString()+" all the way to the edge of the dohyo and out. "+us.toString()+" wins by Oshi-dashi.",false);
};
OpeningMove.Attack = new OpeningMove("Attack");
OpeningMove.Attack.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === OpeningMove.Charge:
			odds = 80; break;
		case theirMove === OpeningMove.Attack:
			odds = 61; break;
		case theirMove === OpeningMove.StepAside:
			odds = 35; break;
		case theirMove === OpeningMove.Slap:
			odds = 57; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for OpeningMove.Attack");
	}
	odds += (us.getStrength() - them.getStrength())/2.5;
	return odds;
};
OpeningMove.Attack.bothSameOpeningMoveNeedingRestart = function(result,us,them) {
	result.addText("Both "+us.toString()+" and "+them.toString()+" go for the belt.",false);
	return false;
};
OpeningMove.Attack.addSuccessfulDescription = function(result,us,them) {
	result.addText(us.toString()+" grabs the belt of "+them.toString()+".");
};
OpeningMove.Attack.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" grabs the belt of "+them.toString()+" and pushes "+them.toString()+" all the way to the edge of the dohyo and out of it. "+us.toString()+" wins by Yori-kiri.",false);
};
OpeningMove.StepAside = new OpeningMove("Step Aside");
OpeningMove.StepAside.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === OpeningMove.Charge:
			odds = 35; break;
		case theirMove === OpeningMove.Attack:
			odds = 80; break;
		case theirMove === OpeningMove.StepAside:
			odds = 61; break;
		case theirMove === OpeningMove.Slap:
			odds = 69; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for OpeningMove.StepAside");
	}
	odds += (us.getAgility() - them.getSpeed())/2.5;
	return odds;
};
OpeningMove.StepAside.bothSameOpeningMoveNeedingRestart = function(_result,_us,_them) {
	return true;
};
OpeningMove.StepAside.addSuccessfulDescription = function(result,us,them) {
	result.addText(us.toString()+" steps to the side and behind "+them.toString()+".");
};
OpeningMove.StepAside.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" steps to the side and behind "+them.toString()+" pushing "+them.toString()+" out of the dohyo. "+us.toString()+" wins by Okuri-dashi.",false);
};
OpeningMove.Slap = new OpeningMove("Slap");
OpeningMove.Slap.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === OpeningMove.Charge:
			odds = 69; break;
		case theirMove === OpeningMove.Attack:
			odds = 57; break;
		case theirMove === OpeningMove.StepAside:
			odds = 46; break;
		case theirMove === OpeningMove.Slap:
			odds = 61; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for OpeningMove.Slap");
	}
	odds += (us.getStrength() - them.getWeight())/2.5;
	return odds;
};
OpeningMove.Slap.bothSameOpeningMoveNeedingRestart = function(result,us,them) {
	result.addText("Both "+us.toString()+" and "+them.toString()+" come out slapping from the tachi-ai.",false);
	return false;
};
OpeningMove.Slap.addSuccessfulDescription = function(result,us,them) {
	result.addText(us.toString()+" slaps "+them.toString()+".");
};
OpeningMove.Slap.addWinDescription = function(result,us,them,theirMove) {
	if (theirMove === OpeningMove.Charge) {
		result.addText(us.toString()+" slaps down "+them.toString()+" as they charge, winning by Hataki-komi.",false);
	} else {
		result.addText(us.toString()+" slaps down "+them.toString()+" winning by Tsuki-dashi.",false);
	}
};

class Continuation {
	constructor(name) {
		this._name = name;
	}
	toString() {
		return this._name;
	}
	getOdds(theirMove,us,them) {
		throw new Error("getOdds not provided for "+this.toString()+".getOdds("+theirMove.toString()+","+us.toString()+","+them.toString()+")");
	}
}
Continuation.Lever = new Continuation("Throw");
Continuation.Lever.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === Continuation.Lever:
			odds = 61; break;
		case theirMove === Continuation.Lift:
			odds = 80; break;
		case theirMove === Continuation.Push:
			odds = 35; break;
		case theirMove === Continuation.Pull:
			odds = 46; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Lever");
	}
	odds += (us.getAgility() - them.getStrength())/2.5;
	return odds;
};
Continuation.Lever.addSuccessfulDescription = function(result,us,them) {
	result.addText(us.toString()+" throws "+them.toString()+".");
};
Continuation.Lever.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" throws "+them.toString()+". "+us.toString()+" wins by Uwate-nage.");
};
Continuation.Lift = new Continuation("Lift");
Continuation.Lift.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === Continuation.Lever:
			odds = 35; break;
		case theirMove === Continuation.Lift:
			odds = 61; break;
		case theirMove === Continuation.Push:
			odds = 80; break;
		case theirMove === Continuation.Pull:
			odds = 57; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Lift");
	}
	odds += (us.getStrength() - them.getWeight())/2.5;
	return odds;
};
Continuation.Lift.addSuccessfulDescription = function(result,us,them) {
	result.addText(us.toString()+" lifts "+them.toString()+".");
};
Continuation.Lift.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(them.toString()+" is lifted out of the dohyo and "+us.toString()+" wins by Yori-taoshi.");
};
Continuation.Push = new Continuation("Push");
Continuation.Push.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === Continuation.Lever:
			odds = 80; break;
		case theirMove === Continuation.Lift:
			odds = 35; break;
		case theirMove === Continuation.Push:
			odds = 59; break;
		case theirMove === Continuation.Pull:
			odds = 69; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Push");
	}
	odds += (us.getWeight() - them.getWeight())/2.5;
	return odds;
};
Continuation.Push.addSuccessfulDescription = function(result,us,them) {
	result.addText(us.toString()+" pushs "+them.toString()+".");
};
Continuation.Push.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(them.toString()+" is pushed out of the dohyo and "+us.toString()+" wins by Oshi-dashi.");
};
Continuation.Pull = new Continuation("Pull");
Continuation.Pull.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === Continuation.Lever:
			odds = 69; break;
		case theirMove === Continuation.Lift:
			odds = 57; break;
		case theirMove === Continuation.Push:
			odds = 46; break;
		case theirMove === Continuation.Pull:
			odds = 59; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Pull");
	}
	odds += (us.getStrength() - them.getWeight())/2.5;
	return odds;
};
Continuation.Pull.addSuccessfulDescription = function(result,us,them) {
	result.addText(us.toString()+" pulls on "+them.toString()+".");
};
Continuation.Pull.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(them.toString()+" is pulled to the clay and "+us.toString()+" wins by Hiki-otoshi.");
};

class Bout {
	static resolve(east,west) {
		const result = new Result();
		Bout._resolveMoves(true,result,east,west);
		while(result.getWinner() === null) {
			Bout._resolveMoves(false,result,east,west);
		}
		return result;
	}
	static _resolveMoves(opening,result,east,west) {
		let eastmove,westmove;
		if (opening) {
			eastmove = east.getOpeningMove(west);
			westmove = west.getOpeningMove(east);
			while(eastmove === westmove && eastmove.bothSameOpeningMoveNeedingRestart(result,east,west)) {
				eastmove = east.getOpeningMove(west);
				westmove = west.getOpeningMove(east);
			}
		} else {
			eastmove = east.getContinuation(west);
			westmove = west.getContinuation(east);
		}

		const eastmovesuccess = (ROT.RNG.getPercentage() <= eastmove.getOdds(westmove,east,west));
		const westmovesuccess = (ROT.RNG.getPercentage() <= westmove.getOdds(eastmove,west,east));
		const eastsaves = east.saves();
		const westsaves = west.saves();
		const eastposswin = eastmovesuccess && !westsaves;
		const westposswin = westmovesuccess && !eastsaves;

		let winner = null, loser = null, winningmove = null, losingmove = null;

		if (eastposswin && westposswin) {
			// Both rikishi succeeded and both failed to save - Need to determine who was the fastest
			const eastspeed = ROT.RNG.getPercentage()+east.getBalance()/2+ROT.RNG.getUniform()/5;
			const westspeed = ROT.RNG.getPercentage()+west.getBalance()/2+ROT.RNG.getUniform()/5;
			if (eastspeed > westspeed) {
				winner = east; loser = west; winningmove = eastmove; losingmove = westmove;
			} else {
				winner = west; loser = east; winningmove = westmove; losingmove = eastmove;
			}
		} else if (eastposswin) {
			winner = east; loser = west; winningmove = eastmove; losingmove = westmove;
		} else if (westposswin) {
			winner = west; loser = east; winningmove = westmove; losingmove = eastmove;
		} else if (eastmovesuccess || westmovesuccess) {
			if (eastmovesuccess) {
				eastmove.addSuccessfulDescription(result,east,west);
			}
			if (westmovesuccess) {
				westmove.addSuccessfulDescription(result,west,east);
			}
			if (opening) {
				result.addText("Both rikishi crash together and lock up.");
			} else {
				result.addText("Both rikishi remain locked up.");
			}
		} else {
			// Neither rikishi succeeded
			if (opening) {
				result.addText("Both rikishi crash together from the tachi-ai and lock up in the middle of the dohyo with neither having the advantage.");
			} else {
				result.addText("Both rikishi remain locked up in the middle of the dohyo with neither having the advantage.");
			}
		}
		if (winner && loser && winningmove) {
			winningmove.addWinDescription(result,winner,loser,losingmove);
			result.setWinner(winner);
			result.setLoser(loser);
		} else if (winner || loser || winningmove) {
			throw new Error("one or more of winner, loser, winningmove set, but not all three!");
		}
	}
}